@@ -28,6 +28,7 @@ sub startup {
 	if($self->mode eq "production") {
 		push @{$self->renderer->paths}, "/usr/share/sreview/templates";
 		push @{$self->static->paths}, "/usr/share/sreview/public";
+		push @{$self->static->paths}, "/usr/share/javascript";
 	} else {
 		push @{$self->static->paths}, "./public";
 		push @{$self->renderer->paths}, "./templates";
@@ -36,6 +37,16 @@ sub startup {
 		push @{$self->static->paths}, $config->get("pubdir");
 	}
 
+	$self->hook(before_dispatch => sub {
+		my $c = shift;
+		my $vpr = $config->get('vid_prefix');
+		my $media = "media-src 'self'";
+		if(defined($vpr) && length($vpr) > 0) {
+			$media = "media-src $vpr;";
+		}
+		$c->res->headers->content_security_policy("default-src 'none'; script-src 'self' 'unsafe-inline'; font-src 'self'; style-src 'self'; img-src 'self'; frame-ancestors 'none'; $media");
+	});
+
 	$self->helper(dbh => sub {
 		state $pg = Mojo::Pg->new->dsn($config->get('dbistring'));
 		return $pg->db->dbh;
@@ -190,7 +201,8 @@ sub startup {
 		$c->stash(comments => $row->{comments});
 		$c->stash(target => "talk_update");
 		$c->stash(layout => 'default');
-		$c->stash(script_raw => 'sreview_viddata = ' . encode_json($viddata) . ';');
+		$c->stash(scripts_raw => ['sreview_viddata = ' . encode_json($viddata) . ';']);
+		$c->stash(scripts_extra => ['/mangler.js']);
 		$c->stash(exten => $config->get('preview_exten'));
 		$c->stash(vid_hostname => $config->get("vid_prefix"));
 	} => 'talk');
@@ -200,7 +212,7 @@ sub startup {
 		my $st;
 		my $conference = {};
 		my $videos = [];
-		my @json = ();
+		my %json;
 		my %formats;
 		my $have_default = 0;
 		$st = $c->dbh->prepare("SELECT MIN(starttime::date), MAX(endtime::date) FROM talks WHERE event = ?");
@@ -212,7 +224,7 @@ sub startup {
 		$st = $c->dbh->prepare("SELECT filename FROM raw_files JOIN talks ON raw_files.room = talks.room WHERE talks.event = ? LIMIT 1");
 		$st->execute($c->eventid);
 		if($st->rows < 1) {
-			$c->render(json => []);
+			$c->render(json => {});
 			return;
 		}
 		$row = $st->fetchrow_hashref;
@@ -230,28 +242,30 @@ sub startup {
 			push @{$conference->{video_formats}}, { $nf => { vcodec => $prof->video_codec, acodec => $prof->audio_codec, resolution => $prof->video_size, bitrate => $prof->video_bitrate } };
 			$formats{$nf} = $prof;
 		}
-		push @json, $conference;
-		$st = $c->dbh->prepare("SELECT title, subtitle, speakerlist(talks.id), description, starttime, starttime::date AS date, to_char(starttime, 'yyyy') AS year, endtime, rooms.name AS room, upstreamid, events.name AS event, slug FROM talks JOIN rooms ON talks.room = rooms.id JOIN events ON talks.event = events.id WHERE state='done' AND event = ?");
+		$json{conference} = $conference;
+		$st = $c->dbh->prepare("SELECT title, subtitle, speakerlist(talks.id), description, starttime, starttime::date AS date, to_char(starttime, 'yyyy') AS year, endtime, rooms.name AS room, rooms.outputname AS room_output, upstreamid, events.name AS event, slug FROM talks JOIN rooms ON talks.room = rooms.id JOIN events ON talks.event = events.id WHERE state='done' AND event = ?");
 		$st->execute($c->eventid);
 		if($st->rows < 1) {
-			$c->render(json => []);
+			$c->render(json => {});
 			return;
 		}
 		my $mt = Mojo::Template->new;
 		$mt->vars(1);
 		while (my $row = $st->fetchrow_hashref()) {
 			my $video = {};
-			my $subtitle = defined($row->{subtitle}) ? $row->subtitle : "";
+			my $subtitle = defined($row->{subtitle}) ? " " . $row->{subtitle} : "";
 			$video->{title} = $row->{title} . $subtitle;
 			$video->{speakers} = [ $row->{speakerlist} ];
 			$video->{description} = $row->{description};
 			$video->{start} = $row->{starttime};
 			$video->{end} = $row->{endtime};
 			$video->{room} = $row->{room};
-			my $outputdir = "";
+			$video->{eventid} = $row->{upstreamid};
+			my @outputdirs;
 			foreach my $subdir(@{$config->get('output_subdirs')}) {
-				$outputdir = join('/', $outputdir, $row->{$subdir});
+				push @outputdirs, $row->{$subdir};
 			}
+			my $outputdir = join('/', @outputdirs);
 			if(defined($config->get('eventurl_format'))) {
 				$video->{details_url} = $mt->render($config->get('eventurl_format'), {
 					slug => $row->{slug},
@@ -262,17 +276,17 @@ sub startup {
 			$video->{video} = join('/',$outputdir, $row->{slug}) . "." . $formats{default}->exten;
 			push @$videos, $video;
 		}
-		push @json, $videos;
-		$c->render(json => \@json);
+		$json{videos} = $videos;
+		$c->render(json => \%json);
 	});
 
 	$r->get('/overview' => sub {
 		my $c = shift;
 		my $st;
 		if($config->get("anonreviews")) {
-			$st = $c->dbh->prepare('SELECT nonce, name, speakers, room, starttime, endtime, state, progress FROM talk_list WHERE eventid = ? AND state IS NOT NULL ORDER BY state, progress, room, starttime');
+			$st = $c->dbh->prepare('SELECT nonce, name, speakers, room, starttime::timestamp, endtime::timestamp, state, progress FROM talk_list WHERE eventid = ? AND state IS NOT NULL ORDER BY state, progress, room, starttime');
 		} else {
-			$st = $c->dbh->prepare('SELECT name, speakers, room, starttime, endtime, state, progress FROM talk_list WHERE eventid = ? AND state IS NOT NULL ORDER BY state, progress, room, starttime');
+			$st = $c->dbh->prepare('SELECT name, speakers, room, starttime::timestamp, endtime::timestamp, state, progress FROM talk_list WHERE eventid = ? AND state IS NOT NULL ORDER BY state, progress, room, starttime');
 		}
 		my $tot = $c->dbh->prepare('SELECT state, count(*) FROM talks WHERE event = ? GROUP BY state ORDER BY state;');
 		my %expls;
@@ -287,6 +301,7 @@ sub startup {
 		$expls{'uploading'} = 'Publishing results';
 		$expls{'done'} = 'Videos published, all done';
 		$expls{'broken'} = 'Review found problems, administrator required';
+		$expls{'ignored'} = 'Talk will not be/was not recorded, ignored for review';
 		$expls{'needs_work'} = 'Fixable problems exist, manual intervention required';
 		$expls{'lost'} = 'Nonfixable problems exist, talk lost';
 		$st->execute($c->eventid) or die;
@@ -316,30 +331,10 @@ sub startup {
 		$c->render;
 	} => 'table');
 
-	$r->post('/talk_update' => sub {
-		my $c = shift;
-		my $nonce = $c->param("nonce");
-		if(!defined($nonce)) {
-			$c->stash(message=>"Unauthorized.");
-			$c->res->code(403);
-			$c->render('error');
-			return undef;
-		}
-		my $sth = $c->dbh->prepare("SELECT id FROM talks WHERE nonce = ? AND state IN ('preview', 'broken')");
-		$sth->execute($nonce);
-		my $row = $sth->fetchrow_arrayref;
-		if(scalar($row) == 0) {
-			$c->stash(message=>"Change not allowed. If this talk exists, it was probably reviewed by someone else while you were doing so too. Please try again later, or check the overview page.");
-			$c->res->code(403);
-			$c->render('error');
-			return undef;
-		}
-		$c->stash(layout => 'default');
-		$c->stash(template => 'talk');
-		$c->flash(completion_message => 'Your change has been accepted. Thanks for your help!');
-		$c->talk_update($row->[0]);
-		$c->redirect_to("/review/$nonce");
-	} => 'talk_update');
+	$r->post('/talk_update')->to(controller => 'talk', action => 'update',
+			layout => 'default', name => 'talk_update',
+			notfound_message => "Unauthorized.",
+			notfound_code => 403);
 
 	my $vol = $r->under('/volunteer' => sub {
 		my $c = shift;
@@ -351,31 +346,7 @@ sub startup {
 		return 1;
 	});
 
-	$vol->get('/list' => sub {
-		my $c = shift;
-		my @talks;
-		$c->dbh->begin_work;
-		my $already = $c->dbh->prepare("SELECT nonce, title, id, state FROM talks WHERE reviewer = ? AND state <= 'preview'");
-		my $new = $c->dbh->prepare("SELECT nonce, title, id, state FROM talks WHERE reviewer IS NULL AND state = 'preview'::talkstate LIMIT ? FOR UPDATE");
-		my $claim = $c->dbh->prepare("UPDATE talks SET reviewer = ? WHERE id = ?");
-		$already->execute($c->session->{id});
-		my $count = $already->rows;
-		if($count < 5) {
-			$new->execute(5 - $count);
-		}
-		for(my $i = 0; $i < $count; $i++) {
-			my $row = [ $already->fetchrow_array ];
-			push @talks, $row;
-		}
-		for(my $i = 0; $i < $new->rows; $i++) {
-			my $row = [ $new->fetchrow_array ];
-			$claim->execute($c->session->{id}, $row->[2]);
-			push @talks, $row;
-		}
-		$c->stash(talks => \@talks);
-		$c->stash(layout => 'admin');
-		$c->dbh->commit;
-	} => 'volunteer/list');
+	$vol->get('/list')->to('volunteer#list');
 
 	my $admin = $r->under('/admin' => sub {
 		my $c = shift;
@@ -397,38 +368,7 @@ sub startup {
 	$admin->any('/schedule/talk/')->to(controller => 'schedule', action => 'mod_talk');
 	$admin->any('/schedule/')->to(controller => 'schedule', action => 'index');
 
-	$admin->get('/' => sub {
-		my $c = shift;
-		my $st;
-		my $talks = ();
-		my $room;
-		my $lastroom = '';
-
-		if(defined($c->session->{room})) {
-			$st = $c->dbh->prepare('SELECT id, room, name, starttime, speakers, state FROM talk_list WHERE eventid = ? AND roomid = ? ORDER BY starttime');
-			$st->execute($c->eventid, $c->session->{room});
-		} else {
-			$st = $c->dbh->prepare('SELECT id, room, name, starttime, speakers, state FROM talk_list WHERE eventid = ? ORDER BY room, starttime');
-			$st->execute($c->eventid);
-		}
-		while(my $row = $st->fetchrow_hashref("NAME_lc")) {
-			if ($row->{'room'} ne $lastroom) {
-				if(defined($room)) {
-					push @$talks, c($lastroom => $room);
-				}
-				$room = [];
-			}
-			$lastroom = $row->{'room'};
-			next unless defined($row->{id});
-			push @$room, [$row->{'starttime'} . ': ' . $row->{'name'} . ' by ' . $row->{'speakers'} . ' (' . $row->{'state'} . ')' => $row->{'id'}];
-		}
-		if(defined($room)) {
-			push @$talks, c($lastroom => $room);
-		}
-		$c->stash(email => $c->session->{email});
-		$c->stash(talks => $talks);
-		$c->render;
-	} => 'admin/main');
+	$admin->get('/')->to('admin#main');
 
 	$admin->get('/logout' => sub {
 		my $c = shift;
@@ -484,7 +424,8 @@ sub startup {
 		$c->stash(comments => $row->{comments});
 		$c->stash(corrections => $viddata);
 		$c->stash(target => "talk_update_admin");
-		$c->stash(script_raw => 'sreview_viddata = ' . encode_json($viddata) . ';');
+		$c->stash(scripts_raw => ['sreview_viddata = ' . encode_json($viddata) . ';']);
+		$c->stash(scripts_extra => ['/mangler.js']);
 		$c->stash(type => "admin");
 		$c->stash(apology => $row->{apologynote});
 		$c->stash(vid_hostname => $config->get("vid_prefix"));
@@ -492,19 +433,10 @@ sub startup {
 		$c->render(template => 'talk');
 	} => 'admin_talk');
 
-	$admin->post('/talk_update' => sub {
-		my $c = shift;
-		my $talk = $c->param("talk");
-		if(!defined($talk)) {
-			$c->stash(message => "Required parameter talk missing.");
-			$c->render("error");
-			return undef;
-		}
-		$c->stash(template => 'talk');
-		$c->flash(completion_message => 'Your change has been accepted. Thanks for your help!');
-		$c->talk_update($talk);
-		$c->redirect_to("/admin/talk?talk=$talk");
-	} => 'talk_update_admin');
+	$admin->post('/talk_update')->to(controller => 'talk', action => 'update',
+			name => 'talk_update_admin',
+			notfound_message => "Required parameter talk missing",
+			notfound_code => 404);
 
 	$admin->get('/brokens' => sub {
 		my $c = shift;
